from peak.rules import when
from env import Env, resolve, extend, add_decl
from type import *
from coercion import *
import report
from astutil.walk import Walker, WalkDown, WalkUp


class Error(Exception):
    pass

class AnalyzerError(Error):
    def __init__(self, what, sloc):
        self.what = what
        self.sloc = sloc
        super(AnalyzerError, self).__init__("semantic analysis failed with: {}".format(what))

def raise_error(message, sloc):
    raise AnalyzerError(message, sloc)


class MapTreeNodes(WalkUp):
    def __init__(self):
        self.classes = {
            'unresolved-type': UnresolvedType,
            'array-type': ArrayType,
            'pointer-type': PointerType,
            'bitvector-type': BitvectorType
        }

    def leave(self, node):
        if node.tag in self.classes:
            return self.classes[node.tag](**node.__dict__)
        return node



def BuiltinEnv():
    types = [
        CharType, ShortType, IntType, LongType,
        UCharType, UShortType, UIntType, ULongType,
        BoolType, VoidType
    ]
    s = Env()
    extend(s, ([e() for e in types]))
    return s


class ResolveTypes(WalkUp):
    def __init__(self):
        self.builtin_env = BuiltinEnv()

    def leave(self, node):
        if node.cls == 'type':
            return node.resolve(self.builtin_env)


def any_function(tag):
    return tag in {'func-decl', 'thread-decl', 'update-decl'}

            
class CollectLabels(WalkDown):
    """Collect all labels into envs for each function. By the way
       generate jump indices for each statement."""
    enter = WalkDown.enter

    @when(enter, "any_function(node.tag)")
    def enter(self, node):
        self.label_env = Env()
        node.label_env = self.label_env

    @when(enter, "node.tag == 'body'")
    def enter(self, node):
        for idx, val in enumerate(node.stmts):
            val.jump_index = idx

    @when(enter, "node.tag == 'label'")
    def enter(self, node):
        self.label_env.add(node.label_name, node)



class ResolveDecls(WalkDown):
    """Collect global env. For each function independently take this
       global env and add param vars to it. Then traverse the function body
       top-down left to right and either resolve symbols or add local declarations
       into the current env."""
    enter = WalkDown.enter

    @when(enter, "node.tag == 'program'")
    def enter(self, node):
        # walk down make sure that root node (program) is entered first
        # add each function, event, thread, update, ... into envs
        xs = ([node.main] + node.funcs + node.threads + node.updates +
              node.events + node.vars)
        self.env = Env()
        extend(self.env, xs)

    @when(enter, "any_function(node.tag)")
    def enter(self, node):
        from astutil.walk import walk_down, AbortWalk
        self.env = Env(self.env)

        node.params = walk_down(node.params, self.enter)
        extend(self.env, node.params)
        self.label_env = node.label_env
        self.func = node
        node.body = walk_down(node.body, self.enter)

        self.env = self.env.parent
        return AbortWalk()

    @when(enter, "node.tag == 'var-decl' and node.loc == 'local'")
    def enter(self, node):
        # add var to current env
        add_decl(self.env, node)

    @when(enter, "node.tag == 'id'")
    def enter(self, node):
        # resolve the declaration for this identifier
        # only var-decl allowed here for now
        decl = resolve(self.env, node.name, node.sloc, {'var-decl'})
        node.decl = decl

    @when(enter, "node.tag == 'call'")
    def enter(self, node):
        # resolve function
        fun = resolve(self.env, node.name, node.sloc, {'func-decl'})
        node.callee = fun

    @when(enter, "node.tag == 'return'")
    def enter(self, node):
        # store reference to enclosing function
        node.func = self.func

    @when(enter, "node.tag == 'goto'")
    def enter(self, node):
        node.label = resolve(self.label_env, node.label_name, node.sloc)
        
    @when(enter, "node.tag == 'resume'")
    def enter(self, node):
        node.thread = resolve(self.env, node.thread_name, node.sloc, {'thread-decl'})

    @when(enter, "node.tag == 'suspend'")
    def enter(self, node):
        node.thread = resolve(self.env, node.thread_name, node.sloc, {'thread-decl'})

    @when(enter, "node.tag == 'notify-event'")
    def enter(self, node):
        node.event = resolve(self.env, node.event_name, node.sloc, {'event-decl'})

    @when(enter, "node.tag == 'wait-event'")
    def enter(self, node):
        node.event = resolve(self.env, node.event_name, node.sloc,  {'event-decl'})

    @when(enter, "node.tag == 'request-update'")
    def enter(self, node):
        node.update = resolve(self.env, node.update_name, node.sloc, {'update-decl'})

    @when(enter, "node.tag == 'dont-initialize'")
    def enter(self, node):
        node.thread = resolve(self.env, node.thread_name, node.sloc, {'thread-decl'})
        node.events = [resolve(self.env, name, node.sloc, {'thread-decl'}) for name in node.event_names]



class TypeCheck(WalkUp):
    """Assume bindings have been established. Now either each leaf
       node has a type (from the bound declaration) or the type can be
       computed from the node itself (e.g. for number or bool nodes).
       Starting from the leaf nodes (bottom-up walk) the types can be
       propagated to the parent nodes."""
    def __init__(self):
        pass

    leave = WalkUp.leave

    def _integer_type(self, value, sloc):
        if value < 0:
            if value >= -0x80000000:
                return IntType()
            if value < -0x8000000000000000:
                report.add_warning("integer range overflow", sloc)
            return LongType()
        else:
            if value < 0x80000000:
                return IntType()
            if value >= 0x8000000000000000:
                report.add_warning("integer range overflow", sloc)
            return LongType()

    @when(leave, "node.tag == 'number'")
    def leave(self, node):
        node.type = self._integer_type(node.value, node.sloc)

    @when(leave, "node.tag == 'id'")
    def leave(self, node):
        node.type = node.decl.type
        node.lvalue = True

    @when(leave, "node.tag == 'bool'")
    def leave(self, node):
        node.type = BoolType()

    @when(leave, "node.tag == 'symbolic'")
    def leave(self, node):
        pass # has already an attribute called 'type' (which is resolved by now)

    @when(leave, "node.tag == 'char'")
    def leave(self, node):
        node.type = CharType()

    @when(leave, "node.tag == 'string'")
    def leave(self, node):
        node.type = ArrayType(inner_type=CharType())

    @when(leave, "node.tag == 'call'")
    def leave(self, node):
        fun = node.callee
        node.type = fun.return_type
        # check that actual arguments match the formal ones
        if len(fun.params) != len(node.args):
            raise_error("expected {} arguments but got {}".format(len(fun.params), len(node.args)), node.sloc)
        node.args = [unpack(e) for e in node.args]

        for a,b in zip(node.args, fun.params):
            a = adapt(a, b.type)

    @when(leave, "node.tag == 'unary'")
    def leave(self, node):
        e = unpack(node.operand)
        if node.op in {'-', '~'}:
            e = expect_bitvector(e)
        elif node.op in {'!'}:
            e = expect_bool(e)
        else:
            assert False
        node.operand = e
        node.type = e.type

    def _handle_pointer_operations(self, node):
        """
         Check if any operand of the binary expression has pointer type.
         If so apply the necessary checks and return a marked result.
        """
        if is_pointer(node.lhs) and is_pointer(node.rhs):
            if not node.op in {'==', '!=', '<', '<=', '>', '>='}:
                raise_error('pointers can only be compared', node.sloc)
            node.type = BoolType()
        elif is_pointer(node.lhs) and is_bitvector(node.rhs):
            if not node.op in {'+', '-'}:
                raise_error('invalid pointer arithmetic', node.sloc)
            node.type = node.lhs.type
        elif is_bitvector(node.lhs) and is_pointer(node.rhs):
            if not node.op in {'+'}: # note: '-' is not supported here
                raise_error('invalid pointer arithmetic', node.sloc)
            node.type = node.rhs.type
            # swap lhs and rhs to be consistent with the above case
            # (pointer is always lhs in pointer arithmetic)
            node.lhs, node.rhs = node.rhs, node.lhs
        elif is_pointer(node.lhs) or is_pointer(node.rhs):
            raise_error('unsupported pointer operation {}'.format(node.op), node.sloc)
        return hasattr(node, 'type')

    @when(leave, "node.tag == 'binary'")
    def leave(self, node):
        lhs = unpack(node.lhs)
        rhs = unpack(node.rhs)
        if self._handle_pointer_operations(node):
            return node
        if node.op in {'&&', '||'}:
            node.lhs = expect_bool(lhs)
            node.rhs = expect_bool(rhs)
        else:
            lhs, rhs = arithmetic_conversion(lhs, rhs)
            node.lhs = expect_bitvector(lhs)
            node.rhs = expect_bitvector(rhs)
        if node.op in {'<', '<=', '>', '>=', '==', '!=', '&&', '||'}:
            node.type = BoolType()
        else:
            node.type = node.lhs.type

    @when(leave, "node.tag == 'cast'")
    def leave(self, node):
        node.expr = unpack(node.expr)
        try:
            node.expr = adapt(node.expr, node.target_type)
        except CoercionError:
            # Implicit conversion to target type failed but explicit
            # conversion currently is possible for all types. Either
            # wrap this expression in specific conversion expressions
            # like 'bitvector-to-pointer', 'pointer-to-pointer', ... or
            # add the necessary logic at runtime. If conversion from and
            # to pointers should not be supported than raise an error here.
            pass

    @when(leave, "node.tag == 'if'")
    def leave(self, node):
        node.cond = unpack(node.cond)
        node.cond = expect_bool(node.cond)

    @when(leave, "node.tag == 'var-decl'")
    def leave(self, node):
        if node.initializer is not None:
            node.initializer = unpack(node.initializer)
            node.initializer = adapt(node.initializer, node.type)

    @when(leave, "node.tag == 'array-type'")
    def leave(self, node):
        if node.length is not None:
            node.length = unpack(node.length)
            node.length = expect_index(node.length)

    @when(leave, "node.tag == 'assign'")
    def leave(self, node):
        if not node.lhs.lvalue:
            raise_error("lvalue expected", node.lhs.sloc)
        node.rhs = unpack(node.rhs)
        node.rhs = adapt(node.rhs, node.lhs.type)

    @when(leave, "node.tag == 'deref'")
    def leave(self, node):
        node.expr = expect_pointer(node.expr)
        node.type = node.expr.type.inner_type
        node.lvalue = True

    @when(leave, "node.tag == 'addrof'")
    def leave(self, node):
        if not node.expr.lvalue:
            raise_error("can only get address of an lvalue expression", node.sloc)
        node.type = PointerType(inner_type=node.expr.type)

    @when(leave, "node.tag == 'access'")
    def leave(self, node):
        node.array = unpack(node.array)
        node.array = expect_pointer(node.array)
        node.index = unpack(node.index)
        node.index = expect_index(node.index)
        node.type = node.array.type.inner_type
        node.lvalue = True

    @when(leave, "node.tag == 'length'")
    def leave(self, node):
        node.array = unpack(node.array)
        node.array = expect_pointer(node.array)
        node.type = ArrayType.index_type


    @when(leave, "node.tag == 'new'")
    def leave(self, node):
        pass # nothing to do here (type is already resolved and processed)

    @when(leave, "node.tag == 'delete'")
    def leave(self, node):
        node.expr = unpack(node.expr)
        node.expr = expect_pointer(node.expr)

        
    @when(leave, "node.tag == 'print'")
    def leave(self, node):
        node.value = unpack(node.value)
        ##node.value = expect_primitive(node.value) # anything can be printed

    @when(leave, "node.tag == 'assume'")
    def leave(self, node):
        node.cond = unpack(node.cond)
        node.cond = expect_bool(node.cond)

    @when(leave, "node.tag == 'assert'")
    def leave(self, node):
        node.cond = unpack(node.cond)
        node.cond = expect_bool(node.cond)

    @when(leave, "node.tag == 'return'")
    def leave(self, node):
        if node.value is not None:
            node.value = unpack(node.value)
            node.value = adapt(node.value, node.func.return_type)

    @when(leave, "node.tag == 'eval'")
    def leave(self, node):
        pass

    @when(leave, "node.tag == 'wait-time'")
    def leave(self, node):
        node.delay = unpack(node.delay)
        node.delay = expect_bitvector(node.delay)

    @when(leave, "node.tag == 'notify-event'")
    def leave(self, node):
        if node.delay:
            node.delay = unpack(node.delay)
            node.delay = expect_bitvector(node.delay)

    @when(leave, "node.tag == 'start-sim'")
    def leave(self, node):
        if node.steps:
            node.steps = unpack(node.steps)
            node.steps = expect_bitvector(node.steps)

    """
     A return stmt is reachable if one is always reachable on each path.
    """
    def _reachable_return(self, block, ip, visited=set()):
        while ip < len(block):
            stmt = block[ip]
            if stmt in visited:
                return True
            visited.add(stmt)
            ip += 1
            if stmt.tag == 'return':
                return True
            if stmt.tag == 'if':
                if not self._reachable_return(block, ip, visited):
                    return False
                ip = stmt.goto.label.jump_index
            if stmt.tag == 'goto':
                ip = stmt.label.jump_index
        return False

    @when(leave, "node.tag == 'func-decl'")
    def leave(self, node):
        if not isa(node.return_type, VoidType()):
            if not self._reachable_return(node.body.stmts, 0):
                raise_error("return stmt is not reachable on all paths but function has non void return type", node.sloc)