from type import *
from astutil.nodes import Expr

def extend_bv(expr, target):
    diff = target.bitwidth - expr.type.bitwidth
    sign = target.signed and expr.type.signed
    assert diff > 0
    return Expr('extend', expr.sloc, type=target, expr=expr, target=target, bitdiff=diff, signed=sign)

def shrink_bv(expr, target):
    diff = expr.type.bitwidth - target.bitwidth
    sign = expr.type.signed and target.signed
    assert diff > 0
    return Expr('slice', expr.sloc, type=target, expr=expr, target=target, bitdiff=diff, signed=sign)


def bitvector_to_bool(expr, target):
    # target is always bool in this case
    return Expr('bitvector-to-bool', expr.sloc, expr=expr, type=BoolType())

def bitvector_to_bitvector(expr, target):
    source = expr.type
    if source.bitwidth < target.bitwidth:
        return extend_bv(expr, target)
    elif source.bitwidth > target.bitwidth:
        return shrink_bv(expr, target)
    else:
        assert source.signed != target.signed
        expr.type = target
        return expr

def bitvector_to_integral(expr, target):
    # target can be any x where 'isa(x, Integral)' is true
    # so check the target type to apply the correct coercion.
    # This coercion can have different ranks depending on target type.
    # Either a promotion or conversion will be applied (depends whether
    # bitwidth will be extended or sliced).
    return bitvector_to_bitvector(expr, target)

def bool_to_bitvector(expr, target):
    return Expr('bool-to-bitvector', expr.sloc, expr=expr, type=target)

def bool_to_integral(expr, target):
    return bool_to_bitvector(expr, target)

def integral_to_integral(expr, target):
    return bitvector_to_bitvector(expr, target)


def expr_or_type(fun):
    def wrapped(arg):
        try:
            type = arg.type
        except AttributeError:
            type = arg
        return fun(type)
    return wrapped

@expr_or_type
def is_bool(type):
    return isa(type, BoolType())

@expr_or_type
def is_bitvector(type):
    return isa(type, BitvectorType)

@expr_or_type
def is_integral(type):
    return isa(type, IntegralType)

@expr_or_type
def is_pointer(type):
    return isa(type, PointerType)

@expr_or_type
def is_array(type):
    return isa(type, ArrayType)

class CoercionError(Exception):
    def __init__(self, what, sloc):
        self.what = what
        self.sloc = sloc
        super(CoercionError, self).__init__("coercion failed with: {}".format(what))

def raise_coercion_failed(expr, target):
    try:
        iter(target)
    except TypeError:
        target = [target]
    s = ' or '.join(("'{}'".format(str(e)) for e in target))
    raise CoercionError("unable to convert '{}' into {}".format(str(expr.type), s), expr.sloc)

def coerce_to(expr, target):
    if isa(expr.type, target):
        return expr
    if is_bitvector(expr) and is_bitvector(target):
        return bitvector_to_bitvector(expr, target)
    if is_bitvector(expr) and is_integral(target):
        return bitvector_to_bitvector(expr, target)
    if is_bitvector(expr) and is_bool(target):
        return bitvector_to_bool(expr, target)
    if is_bool(expr) and is_bitvector(target):
        return bool_to_bitvector(expr, target)
    raise_coercion_failed(expr, target)

def adapt(expr, target):
    return coerce_to(expr, target)


def expect_bitvector(expr):
    if is_bitvector(expr):
        return expr
    if is_bool(expr):
        return coerce_to(expr, IntType())
    raise_coercion_failed(expr, 'bitvector')

def expect_array(expr):
    if not is_array(expr):
        raise_coercion_failed(expr, 'array')
    return expr

def expect_pointer(expr):
    if not is_pointer(expr):
        raise_coercion_failed(expr, 'pointer')
    return expr

def expect_bool(expr):
    return coerce_to(expr, BoolType())

def expect_index(expr):
    return coerce_to(expr, ArrayType.index_type)

def expect_primitive(expr):
    handler = [expect_bitvector, expect_bool]
    for f in handler:
        try:
            return f(expr)
        except CoercionError:
            pass
    raise_coercion_failed(expr, ['bitvector', str(BoolType())])



def assign_conversion(lhs, rhs):
    """
     Called on each assignment. That is direct assign, parameter
     passing and returning function result and variable initializers
     (sort of direct assign). All integer types can be implicitely 
     converted by this function into each other.
    """
    rhs = adapt(rhs, lhs.type)
    return rhs

def integer_promotion(expr):
    """
     Each integer that completely fits into an int should be promoted
     to IntType. Else keep the existing type.
     And if the value is unsigned and does not fit into a signed
     int (has same bitsize) than promote to unsigned int.
     Also promote bool to int.
    """
    if isa(expr.type, (IntType, UIntType)):
        return expr
    if is_bool(expr):
        return Expr('bool-to-integral', expr.sloc, expr=expr, type=IntType())
    if not is_integral(expr.type):
        return expr
    if expr.type.bitwidth < IntType().bitwidth:
        return extend_bv(expr, IntType())
    if expr.type.bitwidth <= UIntType().bitwidth:
        return extend_bv(expr, UIntType())
    return expr


def adapt_bitvectors(lhs, rhs):
    if lhs.type.bitwidth > rhs.type.bitwidth:
        rhs = extend_bv(rhs, lhs.type)
    elif rhs.type.bitwidth > lhs.type.bitwidth:
        lhs = extend_bv(lhs, rhs.type)
    else:
        # equal bitsize
        if lhs.type.signed and rhs.type.signed:
            # unsigned wins (or just let the lhs win as specified in binary op !?)
            #lhs.type.signed = False
            #rhs.type.signed = False
            pass
    return lhs, rhs
    

# [1] (bitvector, bitvector) :
# Extend both operands to the same bitsize and apply the operation.
# Result -> Same type as lhs.

# [2] (bitvector, integral) or (integral, bitvector) :
# Convert integral operand to bitvector (since each integral is a bitvector,
# just keep it as is). Now rule [1] can be applied.

# [3] else :
# Use the C++ operator rules.
# Conversion will be applied on any integral or bool.
def arithmetic_conversion(lhs, rhs):
    if is_bitvector(lhs) and is_bitvector(rhs):
        if not (is_integral(lhs) and is_integral(rhs)):
            return adapt_bitvectors(lhs, rhs)

    """
     Based on C++ Standard 2003, Page 94.
    """
    lhs = integer_promotion(lhs)
    rhs = integer_promotion(rhs)
    types = [ULongType(), LongType(), UIntType()]
    for t in types:
        if isa(lhs.type, t):
            rhs = adapt(rhs, t)
            break
        elif isa(rhs.type, t):
            lhs = adapt(lhs, t)
            break
    return lhs, rhs


def unpack(expr):
    if expr.lvalue:
        rval = Expr('valueof', expr.sloc, expr=expr)
        rval.type = expr.type
        #rval.lvalue = False    # set implicitely for each expression
        return rval
    return expr