import source

"""
required functionality:
 - collect multiple error reports and eventually display them
 - support for multiple levels of error reports (push/pop)
 - perhaps some sort of priority (low significance here)

An error report can contain a header (that might be auto generated
depending on the report group: eg: error, warning, message, ...)
and multiple messages and also multiple source extracts. Messages 
and source extracts should be displayed in the same order they appear
in the report.

A source extract is a range of code based on a source location. The
range can be a range of lines or a range between two source locations.
Extracted code should be prepended by the actual line number and 
all tabs should be expanded. The amount of tab expansion should be
configurable.
"""

class TextNode(object):
    def __init__(self, message, header=None):
        self.message = message
        self.header = header

    def display(self):
        from source import padding as pad
        if self.header != None:
            print ""
            print self.header + ':'
        else:
            print pad
        print pad + self.message
        print pad

class TokenNode(object):
    def __init__(self, token):
        self.token = token

    def display(self):
        from source import padding as pad
        if self.token == None:
            print pad + "token information not available"
            return
        if self.token.source_file == None:
            print pad + "token has no source file"
        else:
            print pad + "file:   " + self.token.source_file
            print pad + "line:   " + str(self.token.line)
            print pad + "column: " + str(self.token.column)
            print pad
            lines = source.format_sloc(self.token, 2)
            for s in lines:
                print s


# collect error (and probably also warnings) reports
class ErrorManager(object):
    def __init__(self):
        self.clear()

    def display(self):
        def display_reports(reports):
            for report in reports:
                for part in report:
                    part.display()

        display_reports(self.reports)
        for level in reversed(self.report_stack):
            for reports in level:
                display_reports(reports)

    def clear(self):
        self.report_stack = []
        self.reports = []
        self.current = None

    def begin_report(self):
        assert self.current == None, "can't open new report until current one is closed"
        self.current = []

    def end_report(self):
        assert self.current != None, "attempt to close a closed report"
        self.reports.append(self.current)
        self.current = None

    def add_part(self, part):
        assert self.current != None, "first open a report"
        self.current.append(part)

    def push_report(self):
        self.report_stack.append(self.reports)
        self.reports = []

    def pop_report(self):
        self.reports = self.report_stack.pop()

    def has_reports(self):
        return len(self.reports) > 0 or len(self.report_stack) > 0

_errormanager = ErrorManager()

### simplified wrapper interface (probably may lead to complete removal of
### the error manager)
_wrapper = [('display_all', _errormanager.display),
            ('clear_all', _errormanager.clear),
            ('push', _errormanager.push_report),
            ('pop', _errormanager.pop_report),
            ('begin', _errormanager.begin_report),
            ('end', _errormanager.end_report),
            ('append', _errormanager.add_part),
            ('available', _errormanager.has_reports)]

for name,func in _wrapper:
    globals()[name] = func

def add_error(what, token):
    begin()
    append(TextNode(what, "error"))
    append(TokenNode(token))
    end()

def display_error(what, token):
    add_error(what, token)
    display_all()
    clear_all()

def add_note(what, token):
    begin()
    append(TextNode(what, "note"))
    append(TokenNode(token))
    end()

def add_warning(what, token):
    begin()
    append(TextNode(what, "warning"))
    apend(TokenNode(token))
    end()

def run_test_case(filename, line, column):
    tok = type('Token', (object,), dict(source_file=filename, text=None, line=line, column=column))
    clear_all()
    add_error("a test error", tok)
    display_all()